/*
This file is part of DashMan video game

Copyright (c) 2008 Michael Georgoulopoulos <msamurai@freemail.gr>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// game.cpp
#include "game.h"
#include "gfx.h"
#include <stdio.h>
#include <vector>
#include <string>

#include "sound.h"

std::vector<Level> levels;
Level curr_level;
int curr_level_id = 0;

// game state
int hero_x = 0;
int hero_y = 0;
Direction hero_dir = DOWN;
int frame = 0;
bool moving = false;
bool moved = false;
bool game_over = false;

int lives = 3;
int score = 0;

class Vec2
{
public:
    int x;
    int y;
    Vec2(int x, int y)
    {
        this->x = x;
        this->y = y;
    }
    void operator += (const Vec2 &v)
    {
        this->x += v.x;
        this->y += v.y;
    }
};

std::vector<Vec2> moving_gems;
std::vector<Vec2> moving_gem_speeds;
std::vector<Vec2> moving_rocks;

static inline GameElement ParseGameElement(char c)
{
    if (c == '#') return WALL;
    if (c == ' ') return SPACE;
    if (c == '.') return SAND;
    if (c == '@') return ROCK;
    if (c == '*') return GEM;
    if (c == '^') return HOME;
    return NONE;
}

static void EnterLevel(int which)
{
    if (which >= (int)levels.size()) {
		return;
	}
    curr_level = levels[which];
    hero_x = curr_level.home_x * 32;
    hero_y = curr_level.home_y * 32;
    moved = false;
    curr_level_id = which;
    PlaySound(SND_NEW_LEVEL);
}

bool LoadLevels()
{
    levels.clear();

    FILE *fp = fopen("data/levels.txt", "rb");
    if (!fp) return false;
    
    Level level;
    int i=0;
    level.gem_count = 0;
    while (!feof(fp))
    {
        char c = getc(fp);
        GameElement el = ParseGameElement(c);
        if (el == NONE) continue;
        if (el == GEM) level.gem_count ++;
        int x = i % 20;
        int y = i / 20;
        if (el == HOME)
        {
            level.home_x = x;
            level.home_y = y;
            printf("%d, %d\n", x, y);
        }
        level.grid[x][y] = el;
        i++;
        if (i == 300)
        {
            levels.push_back(level);
            level.gem_count = 0;
            i = 0;
        }
    }
    
    fclose(fp);
    
    EnterLevel(0);
    
    return true;
}

static int HeroAnim()
{
    int fr = (frame / 5) % 2;
    if (hero_dir == LEFT) return 0 + fr;
    if (hero_dir == UP) return 2 + fr;
    if (hero_dir == DOWN) return 4 + fr;
    if (hero_dir == RIGHT) return 6 + fr;
    return 0;
}

void DrawLevel()
{
    static char tb[1024];
    // draw gameover screen
    if (game_over)
    {
        BlitString(50, 100, 22, "GAME OVER");
        sprintf(tb, "YOUR SCORE IS %d", score);
        BlitString(50, 150, 22, tb);
        BlitString(50, 200, 22, "PRESS ENTER TO CONTINUE");
        return;
    }


	Blit(0, 3, curr_level.home_x*32, curr_level.home_y*32);
    for (int j=0; j<15; j++)
    {
        for (int i=0; i<20; i++)
        {
            int xfr, yfr;
       		bool blit = true;
            switch (curr_level.grid[i][j])
            {
            case WALL:
                xfr = 0; yfr = 2;
                break;
            case SPACE:
                continue;
                break;
            case SAND:
                xfr = 1; yfr = 2;
                break;
            case ROCK:
                xfr = 2; yfr = 2;
                break;
            case GEM:
                xfr = 0; yfr = 1;
                break;
            default:
				blit = false;
            }
            if (blit) Blit(xfr, yfr, i*32, j*32);
        }
    }
    
    // draw hero
    Blit(HeroAnim(), 0, hero_x, hero_y);
    
    // draw moving gems
    for (int i=0; i<(int)moving_gems.size(); i++)
    {
        Blit(0, 1, moving_gems[i].x, moving_gems[i].y);
    }
    
    // draw moving rocks
    for (int i=0; i<(int)moving_rocks.size(); i++)
    {
        Blit(2, 2, moving_rocks[i].x, moving_rocks[i].y);
    }
    
    // draw stats bar
    int bar_pos = 0;
    sprintf(tb, "LIVES %d", lives);
    BlitString(0 + 2, bar_pos + 2, 22, tb, Color(0, 0, 0, 1));
    BlitString(0, bar_pos, 22, tb, Color(1, 1, 1, 1));
    sprintf(tb, "SCORE %d", score);
    BlitString(200 + 2, bar_pos + 2, 22, tb, Color(0, 0, 0, 1));
    BlitString(200, bar_pos, 22, tb, Color(1, 1, 1, 1));
}

void LevelComplete()
{
    if (curr_level_id == (int)levels.size() - 1)
    {
        game_over = true;
        EnterLevel(0);
    }
    else
    {
        EnterLevel(curr_level_id + 1);
    }
}

void ProcessNewCell(int x, int y)
{
    if (curr_level.grid[x][y] == GEM)
    {
		PlaySound(SND_GEM);
        curr_level.grid[x][y] = SPACE;
        // move the gem to the edge of the screen
        moving_gems.push_back(Vec2(x*32, y*32));
        moving_gem_speeds.push_back(
            Vec2((640 - x*32) / 25,
                (0 - y*32) / 25));
    }
    if (curr_level.grid[x][y] == SAND)
    {
        curr_level.grid[x][y] = SPACE;
    }
    
    if (x==curr_level.home_x && y == curr_level.home_y)
    {
        if (!curr_level.gem_count)
            LevelComplete();
    }
}

void Die()
{
	PlaySound(SND_DIE);
    lives--;
    score = 0;
    if (lives > 0)
    {
        hero_x = curr_level.home_x * 32;
        hero_y = curr_level.home_y * 32;
        curr_level = levels[curr_level_id];
        moving_rocks.clear();
        moving_gems.clear();
        moving_gem_speeds.clear();
        moved = false;
    }
    else
    {
        game_over = true;
    }
}

void ProcessRocks()
{
    for (int j=0; j<14; j++)
    {
        for (int i=0; i<20; i++)
        {
            if (curr_level.grid[i][j] == ROCK)
            {
                if ((curr_level.grid[i][j+1] == SPACE ||
				curr_level.grid[i][j+1] == HOME)  &&
                    !((hero_x/32 == i) && (hero_y/32 == j+1)))
                {
                    curr_level.grid[i][j] = SPACE;
                    moving_rocks.push_back(Vec2(i*32, j*32));
                }
            }
        }
    }
    
    for (int i=0; i<(int)moving_rocks.size(); i++)
    {
        // move the rock down
        moving_rocks[i] += Vec2(0, 8);
        if (!(moving_rocks[i].y%32))
        {
            int x = moving_rocks[i].x / 32;
            int y = moving_rocks[i].y / 32;
            if (hero_x/32 == x && hero_y/32 == y)
            {
                Die();
            }
            if (curr_level.grid[x][y+1] != SPACE && curr_level.grid[x][y+1] != HOME)
            {
				PlaySound(SND_ROCK);
                curr_level.grid[x][y] = ROCK;
                moving_rocks.erase(moving_rocks.begin() + i,
                    moving_rocks.begin() + i + 1);
                i--;
            }
            
        }
    }
}

void ProcessGems()
{
    for (int i=0; i<(int)moving_gems.size(); i++)
    {
        moving_gems[i] += moving_gem_speeds[i];
        if (moving_gems[i].y < 0)
        {
            moving_gems.erase(moving_gems.begin() + i,
                moving_gems.begin() + i + 1);
            moving_gem_speeds.erase(moving_gem_speeds.begin() + i,
                moving_gem_speeds.begin() + i + 1);
            i--;
            score += 50 * moving_gems.size() + 50;
            curr_level.gem_count--;
        }
    }
}

bool UpdateGame(Direction user_dir, bool user_action)
{
    if (game_over)
    {
        if (user_action)
        {
            lives = 3;
            score = 0;
            EnterLevel(0);
            game_over = false;
        }
        return true;
    }

    if (user_action && moved)
    {
        Die();
    }
    if (user_dir != NODIR)
    {
        if (!moving)
        {
            int next_x = hero_x / 32;
            int next_y = hero_y / 32;
            switch (user_dir)
            {
            case LEFT:
                next_x -= 1;
                break;
            case RIGHT:
                next_x += 1;
                break;
            case UP:
                next_y -= 1;
                break;
            case DOWN:
                next_y += 1;
                break;
			default:
				break;
            }
            if (curr_level.grid[next_x][next_y] != ROCK &&
                curr_level.grid[next_x][next_y] != WALL)
            {
                moved = true;
                hero_dir = user_dir;
                moving = true;
            }
        }
    }

    int speed = 4;
    if (moving)
    {
        switch (hero_dir)
        {
        case LEFT:
            hero_x -= speed;
            break;
        case RIGHT:
            hero_x += speed;
            break;
        case UP:
            hero_y -= speed;
            break;
        case DOWN:
            hero_y += speed;
            break;
		default:
			break;
        }
        
        if (!(hero_x % 32) && !(hero_y % 32))
        {
            ProcessNewCell(hero_x / 32, hero_y / 32);
            moving = false;
        }
    }
    
    ProcessRocks();
    ProcessGems();
    
    frame++;
    return true;
}
